/*
 * Copyright: Nicholas P Wiebe (2008-2009) and Irmtraud M Meyer (2008-2009)
 * License: licensed under the GNU General Public License version 3 (GPLv3)
 * Main function, runs Transat analysis on all aligned helices.
 */

#include <algorithm>
#include <cassert>
#include <cmath>
#include <ctime>
#include <fstream>
#include <iostream>
#include <string>

#include <sstream>
#include <iomanip>

#include <map>

#include "cmdline.h"
#include "AlignedHelix.h"
#include "Alignment.h"
#include "HelixFinder.h"
#include "Tree.h"
#include "TreeImpl.h"
#include "Utilities.h"
#include "ShuffledAlignment.h"

// TODO: RF00005 seed alignment throws nan with clustalw tree
// TODO: Improvement? Reduce redundant log likelihood calculation with hash-based cache (key is column of bases), gperf for hash + SGI unordered_map
// TODO: BUG? Covariance is 2.00 when conservation is 1.00 in output...?
	// TODO: Check correctness of covariance values produced, seem to conflict with R4RNA
// TODO: dinucleotide shuffling form RNAZ 2.0
// TODO: Profile allHelicesPvalueTable 
// TODO: Have internal tree builder using Bio++
// TODO: incorporate internal MSA (or at least use cmdline MUSCLE?)
// TODO: Unify output standards between TRANSAT and InteRRACT
// Currently TRANSAT outputs to cout, and prints verbose/debug information
// to cerr.  InteRRact on the other hand writes directly to files, and
// prints verbose/debug information to cout

using namespace std;

void printHelp();

bool Tree::nonGapPair = true;
double Tree::rescaleFactor = 1;
bool Alignment::no_cache = false;

int main(int argc, char** argv) {

	gengetopt_args_info args_info;

	if (cmdline_parser(argc, argv, &args_info) != 0) exit(-1);

	time_t seed = time(NULL); // Record for debugging purposes
//	seed = 1316206887; // TODO: fixed random seed is useful for debugging, but this should not be used for predictions. change this to -debug command line parameter?
	srand(seed); // Should be called once and only once in entire program...
//	cout << seed << endl;

	string treeFile = "";
	bool noUnderflow = false;
	bool printLikelihood = false;
	string filename = "";
	string codonfile = "";
	string codonfile2 = "";
	string helixFile = "";

	string outputFilePrefix = "helix";
	string interactionsAlignAFilename = "";
	string interactionsAlignBFilename = "";
	bool interactions = false;
	bool interactions_only = false;
	int shuffle_only = 0;

	string helixLogPrefix = "";

	if (args_info.target_given) filename = args_info.target_arg;
	if (args_info.tree_given) treeFile = args_info.tree_arg;
//	if (args_info.codon_given) codonfile = args_info.codon_arg; // Unsupported
	if (args_info.query_given) {
		interactionsAlignAFilename = args_info.target_arg;
		interactionsAlignBFilename = args_info.query_arg;
		interactions = true;
	}
	if (args_info.output_given) outputFilePrefix = args_info.output_arg;
	if (args_info.min_stem_length_given) Alignment::minStemLength = args_info.min_stem_length_arg;
	if (args_info.min_p_value_given) HelixFinder::pValueRange = args_info.min_p_value_arg;
	if (args_info.no_p_value_given) HelixFinder::doPvalues = false;
//	if (args_info.realign_given) HelixFinder::realign = TCOFFEE; // Unsupported
	if (args_info.no_gap_penalty_given) Tree::nonGapPair = false;
	if (args_info.dot_bracket_given) HelixFinder::print_struct = true;
	if (args_info.subsample_pval_given) HelixFinder::sample_less = true;
	if (args_info.estimate_pval_given) {
		HelixFinder::sample_less = true;
		HelixFinder::integrate = true;
	}
	if (args_info.mantissa_given) noUnderflow = true;
	if (args_info.verbose_given) HelixFinder::debug_verbose = true;
	if (args_info.likelihood_min_given) HelixFinder::logratiothreshold = args_info.likelihood_min_arg;
	if (args_info.p_value_max_given) HelixFinder::pvaluethreshold = args_info.p_value_max_arg;
	if (args_info.likelihood_only_given) printLikelihood = true;
	if (args_info.interactions_only_given) interactions_only = true;
	if (args_info.shuffle_only_given) shuffle_only = args_info.shuffle_only_arg;
	if (args_info.read_helix_given) helixFile = args_info.read_helix_arg;
	if (args_info.tree_factor_given) Tree::rescaleFactor = args_info.tree_factor_arg;
	if (args_info.no_cache_given) Alignment::no_cache = true;

	cmdline_parser_free (&args_info);

	Tree* tree;
	if(noUnderflow) {
		if (HelixFinder::debug_verbose) cerr << "Using BigDouble to prevent numerical underflow in probability calculation" << endl;
		tree = new TreeImpl<BigDouble>(treeFile);
	}
	else {
		if (HelixFinder::debug_verbose) cerr << "Using normal data type (double) for probability calculation. Underflow possible" << endl;
		tree = new TreeImpl<double>(treeFile);
	}

	if (printLikelihood) {
		Alignment align(filename);
		tree->printAllLikelihoods(align);
		delete tree;
		return 0;
	}

	if (shuffle_only > 0) {
		Alignment * alignment = new Alignment(filename);
		vector<string> seqs = alignment->alignedSeqs;
		for (int i = 0; i < shuffle_only; ++i) {
			vector<string> shuffled = ShuffledAlignment::shuffleAlnStatic(seqs, 0);
			stringstream num, ss;
			num << shuffle_only;
			ss << outputFilePrefix << "_" << setfill('0') << setw(num.str().size()) << i + 1 << ".fa";
			ofstream output;
			output.open(ss.str().c_str());
			for (int j = 0; j < seqs.size(); ++j) {
				output << ">" << alignment->seqNames[j] << endl;
				output << shuffled[j] << endl;
			}
		}
		return 0;
	}

	if (interactions) {
		clock_t start = clock();

		if(outputFilePrefix.compare("") == 0) {
			outputFilePrefix = "helices";
		}

		Alignment * alignmentA;
		Alignment * alignmentB;
		alignmentA = new Alignment(interactionsAlignAFilename);
		alignmentB = new Alignment(interactionsAlignBFilename);
		Alignment * concatenatedAlignment = new Alignment(alignmentA, alignmentB);

		if (!Alignment::no_cache ) {
			alignmentA->initializeCache();
			alignmentB->initializeCache();
			concatenatedAlignment->initializeCache();
		}

		map<Tree*, unsigned int> leaf2SeqMap = tree->getLeaf2SeqMap(alignmentA->seqNames);
		assert(leaf2SeqMap.size() == alignmentA->seqNames.size() && leaf2SeqMap.size() == alignmentB->seqNames.size());

		if (!interactions_only) {
			HelixFinder helicesA(alignmentA);
			helicesA.findAllHelices();
			helicesA.writeHelixDataMinimal(outputFilePrefix + "_A.out", tree);

			HelixFinder helicesB(alignmentB);
			helicesB.findAllHelices();
			helicesB.writeHelixDataMinimal(outputFilePrefix + "_B.out", tree);
		}

		HelixFinder helicesAB(concatenatedAlignment);
		
		if (helixFile.empty()) {
			helicesAB.findAllInteractingHelices(*alignmentA, *alignmentB);
		} else {
			helicesAB.readHelix(helixFile);
		}

		helicesAB.writeHelixDataMinimal(outputFilePrefix + "_AB.out", tree);

		delete alignmentA;
		delete alignmentB;
		delete concatenatedAlignment;
	}
	else { // Vanilla Transat (no interactions)
		Alignment * alignment;
		alignment = new Alignment(filename);
		if (!Alignment::no_cache ) {
			alignment->initializeCache();
		}

		//sanity check: make sure all sequences are in tree:
		map<Tree*, unsigned int> leaf2SeqMap = tree->getLeaf2SeqMap(alignment->seqNames);
		assert(leaf2SeqMap.size() == alignment->seqNames.size());
		
		HelixFinder hf(alignment);
		if (helixFile.empty()) {
			hf.findAllHelices();
		} else {
			hf.readHelix(helixFile);
		}
		hf.allHelicesPvalueTable(tree);

		delete alignment;
	}

	delete tree;

	return 0;
}

